package log

import (
	"encoding/base64"
	"fmt"
	"github.com/hpcloud/tail"
	"github.com/open-io/oio-exporter/common"
	"gopkg.in/yaml.v2"
	"log"
	"os"
	"regexp"
	"strconv"
	"strings"
	"time"
)

type daemon struct {
	services **map[string][]*common.Service
	ttl      time.Duration
	patterns map[string]string
	logPfx   string
}

var tailConfig = tail.Config{
	// Logger: tail.DiscardingLogger,
	Follow: true,
	// This prevents channel closure on file moves
	ReOpen: true,
	// This prevents tail from re-parsing previous lines
	Location: &tail.SeekInfo{Offset: 0, Whence: 2},
}

var servicesWithLogFile = []string{"oioswift"}

func parsePatterns(path, defaultPatterns string) (patterns map[string]string, err error) {
	var f *os.File
	f, err = os.Open(path)
	if err != nil {
		if os.IsNotExist(err) {
			var data []byte
			data, err = base64.StdEncoding.DecodeString(defaultPatterns)
			if err != nil {
				return
			}
			err = yaml.Unmarshal(data, &patterns)
			if err != nil {
				return
			}
			f, err = os.OpenFile(path, os.O_CREATE|os.O_WRONLY, 0644)
			if err != nil {
				return
			}
			_, err = f.Write([]byte(data))
			if err != nil {
				return
			}
			f.Close()
			log.Println("Created pattern file", path)
		}
		return
	}
	err = yaml.NewDecoder(f).Decode(&patterns)
	f.Close()
	return
}

func MakeDaemon(patterns, patternPath string, services **map[string][]*common.Service, ttl time.Duration) *daemon {
	logPfx := "/var/log/oio/sds/"

	loadedPatterns, err := parsePatterns(patternPath, patterns)
	log.Printf("INFO: loaded %d log patterns\n", len(loadedPatterns))
	if err != nil {
		log.Println("WARN", "Could not parse patterns", err)
	}

	for svcType, instances := range **services {
		if _, ok := loadedPatterns[svcType]; !ok {
			log.Println("WARN", "Service type", svcType, "has no configured pattern")
			continue
		}
		for _, svc := range instances {
			if _, ok := svc.Conf["log_location"]; !ok {
				parts := strings.Split(svc.ID, "-")
				if len(parts) != 3 {
					log.Println("WARN", "Could not autoparse log location for svc", svc)
					continue
				}
				sfx := ".access"
				if common.Index(servicesWithLogFile, svcType) > -1 {
					sfx = ".log"
				}
				if svcType == "rawx" {
					sfx = "-httpd-access.log"
				}
				svc.Conf["log_location"] = logPfx + strings.Replace(svc.ID, "-", "/", 1) + "/" + parts[1] + "-" + parts[2] + sfx
			}
			svc.Conf["logging"] = "enabled"
		}
	}
	return &daemon{
		services: services,
		ttl:      ttl,
		patterns: loadedPatterns,
		logPfx:   logPfx,
	}
}

func (d *daemon) Run(m *common.MetricData) {
	for _, instances := range **d.services {
		for _, svc := range instances {
			go d.tailLogFile(svc, m)
		}
	}
}

func (d *daemon) tailLogFile(svc *common.Service, m *common.MetricData) {
	if _, ok := svc.Conf["logging"]; !ok {
		return
	}
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
	rTokenize, err := regexp.Compile(d.patterns[svc.Type])
	// detect dot in response time
	multiplier := float64(1)
	for _, v := range strings.Split(d.patterns[svc.Type], " ") {
		if strings.Contains(v, "resp_time") {
			if strings.Contains(v, ".") {
				multiplier = float64(1000)
			}
			break
		}
	}

	path := svc.Conf["log_location"]
	if err != nil {
		m.IncrError("log", svc.Type)
		log.Println("WARN: Fatal error watching file", path, ":", err)
		return
	}
	numTokens := len(rTokenize.SubexpNames())

	for {
		if _, err := os.Stat(path); err != nil {
			m.IncrError("log", svc.Type)
			log.Println("WARN: Could not watch file", path, ":", err, ". Retrying in", d.ttl)
			time.Sleep(d.ttl)
			continue
		}
		t, err := tail.TailFile(path, tailConfig)
		if err != nil {
			m.IncrError("log", svc.Type)
			log.Println("WARN: Could not watch file", path, ":", err, ". Retrying in", d.ttl)
			time.Sleep(d.ttl)
			continue
		}

		for line := range t.Lines {
			if line.Err != nil {
				m.IncrError("log", svc.Type)
				log.Println("WARN: Error watching file", path, ":", err)
				continue
			}
			s := rTokenize.FindStringSubmatch(line.Text)
			if len(s) != numTokens {
				continue
			}
			names := rTokenize.SubexpNames()

			baseLabels := [][]string{
				[]string{"addr", target},
				[]string{"service", svc.Type},
				[]string{"method", s[common.Index(names, "method")]},
			}
			for idx, tag := range names {
				if strings.HasPrefix(tag, "tag_") {
					baseLabels = append(baseLabels, []string{strings.Split(tag, "tag_")[1], s[idx]})
				}
			}
			for idx, name := range names {
				vFloat, err := strconv.ParseFloat(s[idx], 4)
				if err != nil {
					continue
				}
				switch name {
				case "bytes_sent":
					m.Acc(&common.Metric{
						Name:   "log_bandwidth",
						Labels: append(baseLabels, []string{"type", "in"}),
						Value:  vFloat,
						Type:   "stats",
						TTL:    d.ttl,
					})
				case "resp_length":
					m.Acc(&common.Metric{
						Name:   "log_bandwidth",
						Labels: append(baseLabels, []string{"type", "out"}),
						Value:  vFloat,
						Type:   "stats",
						TTL:    d.ttl,
					})
				case "resp_time":
					m.Acc(&common.Metric{
						Name:   "log_response_time",
						Labels: baseLabels,
						Value:  vFloat * multiplier,
						Type:   "stats",
						TTL:    d.ttl,
					})
				case "code":
					m.Acc(&common.Metric{
						Name:   "log_response_code",
						Labels: append(baseLabels, []string{"code", s[common.Index(names, "code")]}),
						Value:  float64(1),
						Type:   "counter",
						TTL:    d.ttl,
					})
				}
			}
		}
	}
}
