package tcp

import (
	"fmt"
	"github.com/open-io/oio-exporter/common"
	"log"
	"strconv"
	"strings"
	"time"
)

type daemon struct {
	services **map[string][]*common.Service
	interval time.Duration
}

func MakeDaemon(services **map[string][]*common.Service, interval time.Duration) *daemon {
	return &daemon{
		services: services,
		interval: interval,
	}
}

func (d *daemon) Run(m *common.MetricData) {
	for serviceType, services := range **d.services {
		switch serviceType {
		case "memcached":
			for _, svc := range services {
				d.collectMemcached(svc, m)
			}
		case "beanstalkd":
			for _, svc := range services {
				d.collectBeanstalkd(svc, m)
			}
		case "zookeeper":
			for _, svc := range services {
				d.collectZookeeper(svc, m)
			}
		}
	}
	time.Sleep(d.interval)
}

func (d *daemon) collectMemcached(svc *common.Service, m *common.MetricData) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)

	data, err := common.TCPCommand(target, "stats\r\n", " ", 1, d.interval)
	if err != nil {
		m.IncrError("tcp", "memcached")
		log.Println("WARN", "in collecting memcached metrics", svc, err)
	}
	for _, v := range data {
		val, err := strconv.ParseFloat(v.Value.(string), 64)
		if err != nil {
			continue
		}
		m.Set(&common.Metric{
			Name: "memcached_" + v.Key,
			Labels: [][]string{
				[]string{"service", target},
				[]string{"service_type", svc.Type},
				[]string{"addr", target},
			},
			Value: val,
		})
	}
}

func (d *daemon) collectZookeeper(svc *common.Service, m *common.MetricData) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)

	data, err := common.TCPCommand(target, "mntr\n", "\t", 0, d.interval)
	if err != nil {
		m.IncrError("tcp", "zookeeper")
		log.Println("WARN", "in collecting zookeeper metrics", svc, err)
	}

	for _, v := range data {
		val, err := strconv.ParseFloat(v.Value.(string), 64)
		if err != nil {
			continue
		}
		m.Set(&common.Metric{
			Name: v.Key,
			Labels: [][]string{
				[]string{"service", target},
				[]string{"service_type", svc.Type},
				[]string{"addr", target},
			},
			Value: val,
		})
	}
}

func (d *daemon) collectBeanstalkd(svc *common.Service, m *common.MetricData) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)

	data, err := common.TCPCommand(target, "stats\r\n", ": ", 0, d.interval)
	if err != nil {
		m.IncrError("tcp", "beanstalkd")
		log.Println("WARN", "in collecting beanstalkd metrics", svc, err)
	}

	for _, v := range data {
		val, err := strconv.ParseFloat(v.Value.(string), 64)
		if err != nil {
			continue
		}
		m.Set(&common.Metric{
			Name: "beanstalkd_" + strings.Replace(v.Key, "-", "_", -1),
			Labels: [][]string{
				[]string{"service", target},
				[]string{"service_type", svc.Type},
				[]string{"addr", target},
			},
			Value: val,
		})
	}

	tubes, err := common.TCPCommand(target, "list-tubes\r\n", " ", 0, d.interval)
	if err != nil {
		m.IncrError("tcp", "beanstalkd")
		log.Println("WARN", "in listing beanstalkd tubes", svc, err)
	}

	for _, tube := range tubes {
		data, err := common.TCPCommand(target, "stats-tube "+tube.Value.(string)+"\r\n", ": ", 0, d.interval)
		if err != nil {
			m.IncrError("tcp", "beanstalkd")
			log.Println("WARN", "in collecting beanstalkd metrics for tube", tube.Value.(string), svc, err)
			continue
		}
		for _, v := range data {
			val, err := strconv.ParseFloat(v.Value.(string), 64)
			if err != nil {
				continue
			}
			m.Set(&common.Metric{
				Name: "beanstalkd_" + strings.Replace(v.Key, "-", "_", -1),
				Labels: [][]string{
					[]string{"service", target},
					[]string{"service_type", svc.Type},
					[]string{"addr", target},
					[]string{"tube", tube.Value.(string)},
				},
				Value: val,
			})
		}
	}
}
